﻿% -----------------------------------------------------------------------------
% File: check_spl.pl
% Content: SPL checks for FLAME
% Author: Amador Durán Toro
% Version: 2.0
% Date: 19/03/2015 (Día del Padre)
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% Z definitions (AFL)
%   +- SPL ------------------------------------
%   | model : Model
%   | features : F1 Feature
%   +------------------
%   | features( model ) = features
%   | [ characteristic model checks ]
%   +------------------------------------------
%
% Prolog predicates
%   check_spl( spl( F, M ) )
%
% This predicate checks an SPL for internal correctness:
%   - if its set of features is not empty
%   - if its set of features is actually a set
%   - if its model is defined using only its set of features
%   - if its characteristic model is correct, using the abstract predicate
%     checkModel, that must redefined in the characteristic model layer.
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% check_spl( SPL )
% It checks an SPL for internal correctnes.
% -----------------------------------------------------------------------------

% No features
check_spl( spl( [], _ ) ) :-
   write_error( 'The feature set is empty.' ),
   fail.

% Features is not a set, i.e. it contains duplicates
check_spl( spl( F, _ ) ) :-
   not( is_set( F ) ),
   write_error( 'The feature set contains duplicates.' ),
   fail.

% Feature model uses unknown features
check_spl( spl( F, M ) ) :-
   features( M, MF ),
   subtract( MF, F, U ),
   U \= [],
   write_error( [ 'Characteristic model includes unknown features: ', U ] ),
   fail.

% Some features are not used in the feature model
check_spl( spl( F, M ) ) :-
   features( M, MF ),
   subtract( F, MF, U ),
   U \= [],
   write_error( [ 'Characteristic model does not use some features: ', U ] ),
   fail.

% Check characteristic model
check_spl( spl( F, M ) ) :-
   check_model( spl( F, M ) ), % abstract predicate
   fail.

% Success predicate (nothing else to be checked)
check_spl( spl( _, _ ) ) :-
  write( 'No (more) errors found in SPL.' ), nl.

% -----------------------------------------------------------------------------
% write_error( [ M ] )
% It writes a (list of) messages as an error of the SPL.
% -----------------------------------------------------------------------------

write_error( M ) :-
   is_list( M ), !,
   write( 'SPL is not OK! -> ' ),
   findall( Mi, ( member( Mi, M ), write( Mi ) ), _ ), nl.

write_error( Mi ) :-
   write( 'SPL is not OK! -> ' ), write( Mi ), nl.
