﻿% -----------------------------------------------------------------------------
% File: sets.pl
% Content: Set toolkit for FLAME
% Author: Amador Durán Toro
% Version: 2.0
% Date: 19/03/2015
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% Prolog predicates (public)
%   equal_set( [ Set ], [ Set ] )
%   equal_superset( [ [ Set ] ], [ [ Set ] ] )
%   union( [ [ Set ] ], [ Set ] )
%   intersection( [ [ Set ] ], [ Set ] )
%   maximal( [ Set ], [ [ Set ] ] )
%   proper_subset( [ Subset ], [ Set ] )
%   subset_or_disjoint( [ Subset ], [ Set ] )
%   power_set( [ Set ], [ [ Powerset ] ] )
%   safe_sort( [ Set ], [ Set ] )
%   for_each_union( Goal, [ Set ], [ Set ] )
%   for_each_sort( [ [ Set ] ], [ [ Set ] ] )
%
% Some of these set-related predicates are implemented in other versions of
% Prolog, but not in SWI-Prolog.
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% equal_set( [ Set ], [ Set ] )
% Predicate that checks set equality as mutual inclusion.
% -----------------------------------------------------------------------------

equal_set( Set1, Set2 ) :-
  subset( Set1, Set2 ),
  subset( Set2, Set1 ).

% -----------------------------------------------------------------------------
% equal_superset( [ [ Set ] ], [ [ Set ] ] )
% Predicate that checks superset (set of sets) equality.
% -----------------------------------------------------------------------------

equal_superset( S1, S2 ) :-
  for_each_sort( S1, SS1 ),
  for_each_sort( S2, SS2 ),
  equal_set( SS1, SS2 ).

% -----------------------------------------------------------------------------
% union( [ [ Set ] ], [ Set ] )
% It computes the distributed union over a set of sets.
% -----------------------------------------------------------------------------

union( Superset, Union ) :-
  flatten( Superset, Flatten ),
  list_to_set( Flatten, Union ).

% -----------------------------------------------------------------------------
% intersection( [ [ Set ] ], [ Set ] )
% It computes the distributed intersection over a set of sets using the
% auxiliary predicate dinter.
% -----------------------------------------------------------------------------

intersection( [], [] ).

intersection( [ S1 | Rest ], Intersection ) :-
  dinter( [ S1 | Rest ], S1, Intersection ).

dinter( [], Result, Result ).

dinter( [ S1 | Rest ], PreviousResult, Result ) :-
  intersection( S1, PreviousResult, Aux ),
  dinter( Rest, Aux, Result ).

% -----------------------------------------------------------------------------
% maximal( [ Set ], [ [ Set ] ] )
% This backtrackable predicate checks if the first set is a maximal subset of
% the set of sets passed as the second argument.
% -----------------------------------------------------------------------------

maximal( Si, S ) :-
  member( Si, S ),
  forall( member( Sj, S ), not( proper_subset( Si, Sj ) ) ).

% -----------------------------------------------------------------------------
% proper_subset( [ Subset ], [ Set ] )
% It checks if the first set is a proper subset of the second. In order to
% ignore if both sets are sorted, lengths are compared to determine if they
% are not the same set.
% -----------------------------------------------------------------------------

proper_subset( Si, Sj ) :-
  length( Si, L1 ),
  length( Sj, L2 ),
  L1 \= L2,
  subset( Si, Sj ).

% -----------------------------------------------------------------------------
% proper_subset_superset( [ Subset_superset ], [ Set_superset ] )
% It checks if the first superset is a proper subset of the second. In order to
% ignore if both sets are sorted, lengths are compared to determine if they
% are not the same set.
% -----------------------------------------------------------------------------

proper_subset_superset( Si, Sj ) :-
  length( Si, L1 ),
  length( Sj, L2 ),
  L1 \= L2,
  for_each_sort( Si, SS1 ),
  for_each_sort( Sj, SS2 ),
  subset( SS1, SS2 ).

% -----------------------------------------------------------------------------
% disjoint( [ Set ], [ Set ] )
% It checks if the two sets are disjoint.
% -----------------------------------------------------------------------------

disjoint( S1, S2 ) :-
  intersection( S1, S2, [] ).

% -----------------------------------------------------------------------------
% power_set( [ Set ], [ [ Powerset ] ] )
% Powerset predicate. It simply reverses the set before computing its powerset
% in order to generate more "human-like" results (first element first, etc.).
% -----------------------------------------------------------------------------

power_set( Set, Powerset ) :-
  reverse( Set, SetReverse ), !,
  power_set_aux( SetReverse, [[]], Powerset ), !.

% -----------------------------------------------------------------------------
% power_set_aux( [ Set ], [ Acc ], [ Powerset ] )
% Helper predicates borrowed from "Simply Logical" from Peter Flach (2005)
% http://www.cs.bris.ac.uk/Teaching/Resources/COMS30106/slides/SLchapter3.pdf
% -----------------------------------------------------------------------------

power_set_aux( [], Powerset, Powerset ).

power_set_aux( [ H | T ], Acc, PowerSet ):-
  extend_pset( H, Acc, Acc1 ), 
  power_set_aux( T, Acc1, PowerSet ).

extend_pset( _, [], []).

extend_pset( H, [ List | MoreLists ],[ List, [ H | List ] | More ] ):-
  extend_pset( H, MoreLists, More ).

% -----------------------------------------------------------------------------
% for_each_sort( [ [ Set ] ], [ [ Set ] ] )
% It internally sorts the sets in the first superset into the second superset.
% -----------------------------------------------------------------------------

for_each_sort( X, Y ) :-
  is_list( X ),
  aggregate_all( bag( Yi ), ( member( Xi, X ), sort( Xi, Yi ) ), Y ).

% for_each_sort( [], [] ).
% 
% for_each_sort( [ Xi | X ], [ Yi | Y ] ) :-
%   sort( Xi, Yi ),
%   for_each_sort( X, Y ).

% -----------------------------------------------------------------------------
% for_each_union( Goal, [ Set ], [ Set ] )
% It applies the goal of the form Goal( In, Out ) or Goal( K, In, Out ) to all
% the members of the first set and stores the results in the second set.
% -----------------------------------------------------------------------------

for_each_union( _, [], [] ).

for_each_union( G, [ Xi | X ], Y ) :-
  call( G, Xi, Y1 ),
  for_each_union( G, X, Y2 ),
  union( Y1, Y2, Y ).

for_each_union( _, _, [], [] ).

for_each_union( G, K, [ Xi | X ], Y ) :-
  call( G, K, Xi, Y1 ),
  for_each_union( G, K, X, Y2 ),
  union( Y1, Y2, Y ).
