﻿% -----------------------------------------------------------------------------
% File: check_model.pl
% Content: BFM checks for FLAME (refactored)
% Author: Amador Durán Toro
% Version: 3.0
% Date: 19/03/2015 (Día del Padre)
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% Z definitions (CML/BFM)
%   +- SPL ------------------------------------
%   | model : Model
%   | features : F1 Feature
%   +------------------
%   | features( model ) = features
%   |
%   | forAll ctc_i : ctc model @
%   |   features( ctc_i ) subseteq features( tree model )
%   |
%   | ran( feature_bag( tree model ) ) = { 1 }
%   +------------------------------------------
%
% Prolog predicates
%   check_model( spl( F, bfm( TREE, CTC ) ) )
%   check_ctcs( bfm( TREE, CTC ) )
%   check_tree( TREE )
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% check_model( spl( _, bfm( TREE, CTC ) ) )
% It checks if all CTCs use features in the tree
% and if the tree is actually a tree
% -----------------------------------------------------------------------------

check_model( spl( _, bfm( TREE, CTC ) ) ) :-
   check_ctcs( bfm( TREE, CTC ) ),
   check_tree( TREE ).

% -----------------------------------------------------------------------------
% check_ctcs( bfm( TREE, CTC ) )
% It checks if all CTCs use different features and their features are in the tree.
% -----------------------------------------------------------------------------

check_ctcs( bfm( _, [] ) ) :- !.

check_ctcs( bfm( TREE, CTC ) ) :-
  check_same_feature( CTC ),
  check_tree_features( bfm( TREE, CTC ) ).
  
check_same_feature( CTC ) :-
  findall( CTCi,
    ( member( CTCi, CTC ), x_features( CTCi, XFi ), length( XFi, 1 ) ), RESULT ),
  RESULT \= [], !,
  write_error( [
    'Some CTCs are defined over the same feature: ', RESULT ] ).

check_same_feature( _ ).

check_tree_features( bfm( TREE, CTC ) ) :-
  t_features( TREE, F1 ),
  for_each_union( x_features, CTC, F2 ),
  subtract( F2, F1, RESULT ),
  RESULT \= [], !,
  write_error( [
    'Some CTCs include features that do not appear in the feature tree: ', RESULT ] ).
    
check_tree_features( bfm( _, _ ) ).

% -----------------------------------------------------------------------------
% check_tree( TREE )
% It checks if the features appear only once in the feature tree.
% -----------------------------------------------------------------------------

check_tree( TREE ) :-
  t_feature_bag( TREE, FB ),
  bag_range( FB, R ),
  R \= [ 1 ], !,
  bag_to_pairs( FB, FP ),
  findall( Fi-Ni, ( member( Fi-Ni, FP ), Ni > 1 ), RESULT ),
  write_error( [
    'Some features appear more than once in the feature tree: ', RESULT ] ).

check_tree( _ ).
