﻿% -----------------------------------------------------------------------------
% File: instance.pl
% Content: BFM is-instance-of relation for FLAME
% Author: Amador Durán Toro
% Version: 2.0
% Date: 04/03/2015
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% NOTE: instance is a built-in predicate of SWI-Prolog that cannot redefined
% -----------------------------------------------------------------------------
%
% Z definitions
%   _instance_ : Product <-> Model
%   _tinstance_ : Product <-> TreeFeature
%   _rinstance_ : Product <-> Relationship
%   _xinstance_ : Product <-> CTC
%
% Prolog predicates
%   instance_of( Product, Model )
%   instance_of_t( Product, TreeFeature )
%   instance_of_r( Product, Relationship )
%   instance_of_x( Product, CTC )
%   instance_or_disjoint( Product, TreeFeature, NOI )
%   instance_or_disjoint( Product, [ TreeFeature ], NOI )
%
% The first predicates check if a product is an instance of a BFM model, a tree,
% a relationship, and a CTC. The last one is a helper predicate that checks if a
% product is an instance of or does not contain any feature of a list of trees.
% It also computes the number of times the product is an instance of the treee
% (NOI = Number Of Instances).
% -----------------------------------------------------------------------------

% -----------------------------------------------------------------------------
% instance_of( Product, Model )
% It computes if the product is an instance of a BFM model.
% -----------------------------------------------------------------------------

instance_of( P, bfm( TREE, CTC ) ) :-
  instance_of_t( P, TREE ), !,
  forall( member( CTCi, CTC ), instance_of_x( P, CTCi ) ).

% -----------------------------------------------------------------------------
% instance_of_t( Product, TreeFeature )
% It computes if the product is an instance of a TreeFeature.
% -----------------------------------------------------------------------------

instance_of_t( P, feature( F, R ) ) :-
  member( F, P ),
  forall( member( Ri, R ), instance_of_r( P, Ri ) ).

% -----------------------------------------------------------------------------
% instance_of_r( Product, Relationship )
% It computes if the product is an instance of a Relationship.
% -----------------------------------------------------------------------------

instance_of_r( P, mandatory( Ti ) ) :-
  instance_of_t( P, Ti ).

instance_of_r( P, optional( Ti ) ) :-
  instance_or_disjoint( P, Ti, _ ).

instance_of_r( P, one_or_more( T ) ) :-
  instance_or_disjoint( P, T, NOI ),
  NOI >= 1.

instance_of_r( P, only_one( T ) ) :-
  instance_or_disjoint( P, T, NOI ),
  NOI =:= 1.

% -----------------------------------------------------------------------------
% instance_of_x( Product, CTC )
% It computes if the product is an instance of a CTC.
% -----------------------------------------------------------------------------

instance_of_x( P, requires( F1, F2 ) ) :-
  not( member( F1, P ) )
  ;
  member( F2, P ).

instance_of_x( P, excludes( F1, F2 ) ) :-
  not( member( F1, P ) )
  ;
  not( member( F2, P ) ).

% -----------------------------------------------------------------------------
% instance_or_disjoint( Product, [ TreeFeature ], NOI )
% instance_or_disjoint( Product, TreeFeature, NOI )
%
% It checks if the product is an instance of the TreeFeature or if none of the
% features in the product appear in the tree. It also computes how many times
% the product is an instance of the tree (NOI = Number Of Instances).
% -----------------------------------------------------------------------------

instance_or_disjoint( _, [], NOI ) :-
  NOI is 0.

instance_or_disjoint( P, [ TFi | TF ], NOI ) :-
  instance_or_disjoint( P, TFi, NOI1 ),
  instance_or_disjoint( P, TF, NOI2 ),
  NOI is NOI1 + NOI2, !.

instance_or_disjoint( P, TFi, NOI ) :-
  instance_of_t( P, TFi ),
  NOI is 1.

instance_or_disjoint( P, TFi, NOI ) :-
  t_features( TFi, F ),
  intersection( P, F, [] ),
  NOI is 0.
